#ifndef IMAGE_H
#define IMAGE_H
#include <algorithm>
#include <fstream>
#include <iostream>
#include <string>
#include <sstream>

typedef unsigned int uint;
typedef unsigned char uchar;

class Image
{
    uint width, height;
    uchar *data;

public:

    inline Image () : width (0), height (0), data (0)
    {
    }

    inline Image (uint width, uint height) : width (width), height (height), data (getDim () ? new uchar[getDim ()] : 0)
    {
        std::fill (data, data + getDim (), 0);
    }

    inline Image (const Image & other) : width (other.width), height (other.height), data (getDim () ? new uchar[getDim ()] : 0)
    {
        std::copy (other.data, other.data + getDim (), data);
    }

    inline ~Image ()
    {
        if ( data ) delete[] data;
    }

    inline uint getDim () const
    {
        return width * height;
    }

    inline uint getWidth () const
    {
        return width;
    }

    inline uint getHeight () const
    {
        return height;
    }

    inline uchar & operator () (int x, int y)
    {
        return data[x + width * y];
    }

    inline const uchar & operator () (int x, int y) const
    {
        return data[x + width * y];
    }

    inline Image & operator= (const Image & other)
    {
        if ( width != other.width || height != other.height )
        {
            if ( data ) delete[] data;
            width = other.width;
            height = other.height;
            data = getDim () ? new uchar[getDim ()] : 0;
        }
        if ( this != &other )
            std::copy (other.data, other.data + getDim (), data);
        return *this;
    }

    inline Image operator+(const Image &other) const
    {
        Image tmp(std::max(width,other.width),std::max(height,other.height) );

        for ( uint i = 0; i < width; ++i )
        {
            for ( uint j = 0; j < height; ++j )
            {
                tmp(i,j) = ( *this )( i,j );
            }
        }

        for ( uint i = 0; i < other.width; ++i )
        {
            for ( uint j = 0; j < other.height; ++j )
            {
                tmp(i,j) = std::min(tmp(i,j) + other(i,j),255);
            }
        }
        return tmp;
    }

    inline Image & operator+=(const Image &other)
    {
        return *this = *this + other;
    }


    inline Image operator-(const Image &other) const
    {
        Image tmp(std::max(width,other.width),std::max(height,other.height) );

        for ( uint i = 0; i < width; ++i )
        {
            for ( uint j = 0; j < height; ++j )
            {
                tmp(i,j) = ( *this )( i,j );
            }
        }

        for ( uint i = 0; i < other.width; ++i )
        {
            for ( uint j = 0; j < other.height; ++j )
            {
                if ( tmp(i,j) <= other(i,j) )
                    tmp(i,j) = 0;
                else
                    tmp(i,j) -= other(i,j);
            }
        }
        return tmp;
    }

    inline Image & operator-=(const Image &other)
    {
        return *this = *this - other;
    }

    static void saveImage (const Image * image, const char *filename)
    {
        if ( !image || !image->getDim () ) return;
        std::ofstream s (filename);
        s << "P2" << std::endl << image->width << " " << image->height << std::endl << (int) *std::max_element (image->data,
                                                                                                                image->data +
                                                                                                                image->
                                                                                                                getDim () ) << std::endl;
        for ( uint y = 0; y < image->height; ++y )
        {
            for ( uint x = 0; x < image->width; ++x )
            {
                s.width (4);
                s << (int) ( *image )( x, y ) << std::right;
            }
            s << std::endl;
        }
    }

    static Image *loadImage (const char *filename)
    {
        std::ifstream s (filename);

        std::string a;

        getline (s, a);
        if ( a != "P2" )
            return 0;
        while ( getline (s, a) )
        {
            if ( a.substr (0, 1) != "#" )
                break;
        }
        int w,h;
        int maxgreyvalue;

        std::stringstream strstr1 (a);
        strstr1 >> w;
        strstr1 >> h;

        getline (s, a);
        std::stringstream strstr2 (a);
        strstr2 >> maxgreyvalue;

        Image *pic = new Image (w, h);

        uchar * dataptr = pic->data;
        while ( getline (s, a) )
        {
            std::istringstream istr (a);
            while ( istr.good () )
            {
                int i;
                istr >> i;
                *dataptr++ = i;
            }
        }
        return pic;
    }
};

#endif
